==> llm2vec-0.1.0-py2.py3-none-any.whl: llm2vec/llm2vec.py <==
import logging
from functools import partial
import numpy as np
import torch
from torch import nn, Tensor, device
import torch.multiprocessing as mp
from typing import Dict, List, Union, Optional
from tqdm.autonotebook import trange
from transformers import AutoModel, AutoTokenizer, LlamaConfig, MistralConfig

logger = logging.getLogger(__name__)

def batch_to_device(batch, target_device: device):
    """
    send a pytorch batch to a device (CPU/GPU)
    """
    for key in batch:
        if isinstance(batch[key], Tensor):
            batch[key] = batch[key].to(target_device)
    return batch


class LLM2Vec(nn.Module):
    def __init__(
        self,
        model: AutoModel,
        tokenizer: AutoTokenizer,
        pooling_mode: str = "mean",
        max_length: int = 512,
        doc_max_length: int = 400,
        skip_instruction: bool = True
    ):
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.pooling_mode = pooling_mode
        self.skip_instruction = skip_instruction
        self.max_length = max_length
        self.doc_max_length = doc_max_length

    def prepare_for_tokenization(self, text):
        def _is_instruct(name):
            return ("chat" in name.lower()) or ("instruct" in name.lower()) or ("sharegpt" in name.lower())
        
        if _is_instruct(self.model.config._name_or_path):
            text = '[INST] ' + text.strip() + ' [/INST]'
        if (isinstance(self.model.config, LlamaConfig) or isinstance(self.model.config, MistralConfig)) and self.pooling_mode == "eos_token":
            text = text.strip() + ' </s>'
        return text

    def tokenize(self, texts):
        texts_2 = []
        original_texts = []
        for text in texts:
            t = text.split("!@#$%^&*()")
            texts_2.append(t[1])
            original_texts.append("".join(t))
        
        original = self.tokenizer(original_texts, return_tensors='pt', padding=True, truncation=True, max_length=self.max_length)
        embed_mask = None
        for t_i, t in enumerate(texts_2):
            ids = self.tokenizer([t], return_tensors='pt', padding=True, truncation=True, max_length=self.max_length, add_special_tokens=False)
            if embed_mask is None:
                e_m = torch.zeros_like(original["attention_mask"][t_i])
                if len(ids["input_ids"][0]) > 0:
                    e_m[-len(ids["input_ids"][0]):] = torch.ones(len(ids["input_ids"][0]))
                embed_mask = e_m.unsqueeze(0)
            else:
                e_m = torch.zeros_like(original["attention_mask"][t_i])
                if len(ids["input_ids"][0]) > 0:
                    e_m[-len(ids["input_ids"][0]):] = torch.ones(len(ids["input_ids"][0]))
                embed_mask = torch.cat((embed_mask, e_m.unsqueeze(0)), dim=0)

        original["embed_mask"] = embed_mask
        return original

    def _skip_instruction(self, sentence_feature):
        assert sentence_feature["attention_mask"].shape == sentence_feature["embed_mask"].shape
        sentence_feature["attention_mask"] = sentence_feature["embed_mask"]

    def forward(self, sentence_feature: Dict[str, Tensor]):
        embed_mask = None
        if "embed_mask" in sentence_feature:
            embed_mask = sentence_feature.pop("embed_mask")
        reps = self.model(**sentence_feature)
        sentence_feature["embed_mask"] = embed_mask
        
        return self.get_pooling(sentence_feature, reps.last_hidden_state)
    
    def get_pooling(self, features, last_hidden_states):  # All models padded from left
        assert self.tokenizer.padding_side == 'left', "Pooling modes are implemented for padding from left."
        if self.skip_instruction:
            self._skip_instruction(features)
        seq_lengths = features["attention_mask"].sum(dim=-1)
        if self.pooling_mode == "mean":
            return torch.stack([last_hidden_states[i, -length:, :].mean(dim=0) for i, length in enumerate(seq_lengths)], dim=0) 
        elif self.pooling_mode == "weighted_mean":
            bs, l, _ = last_hidden_states.shape
            complete_weights = torch.zeros(bs, l, device=last_hidden_states.device)
            for i, seq_l in enumerate(seq_lengths):
                if seq_l > 0:
                    complete_weights[i, -seq_l:] = torch.arange(seq_l) + 1
                    complete_weights[i] /= torch.clamp(complete_weights[i].sum(), min=1e-9)
            return torch.sum(last_hidden_states * complete_weights.unsqueeze(-1), dim=1)
        elif self.pooling_mode == "eos_token" or self.pooling_mode == "last_token":
            return last_hidden_states[:, -1]
        elif self.pooling_mode == "bos_token":
            return last_hidden_states[features["input_ids"]==self.tokenizer.bos_token_id]
        else:
            raise ValueError(f"{self.pooling_mode} is not implemented yet.")

    def _convert_to_str(self, instruction, text):   
        tokenized_q = self.tokenizer(text, return_tensors='pt', padding=True, truncation=True, max_length=self.max_length, add_special_tokens=False)
        tokenized_q_length = len(tokenized_q["input_ids"][0])

        while tokenized_q_length > self.doc_max_length:
            reduction_ratio = self.doc_max_length / tokenized_q_length
            reduced_length = int(len(text.split()) * reduction_ratio)
            text = " ".join(text.split()[:reduced_length])
            tokenized_q = self.tokenizer(text, return_tensors='pt', padding=True, truncation=True, max_length=self.max_length, add_special_tokens=False)
            tokenized_q_length = len(tokenized_q["input_ids"][0])
    
        return f"{instruction.strip()} !@#$%^&*(){text}"

    def encode(self, sentences: Union[str, List[str]],
               batch_size: int = 32,
               show_progress_bar: bool = True,
               convert_to_numpy: bool = True,
               convert_to_tensor: bool = False,
    ):
        if isinstance(sentences[0],str) and isinstance(sentences[-1],int):
            sentences = [sentences]
        # required for MEDI version of MTEB
        if isinstance(sentences[0],list):
            concatenated_input_texts = []
            for sentence in sentences:
                assert isinstance(sentence[0], str)
                assert isinstance(sentence[1], str)
                concatenated_input_texts.append(self._convert_to_str(sentence[0], sentence[1]))
            sentences = concatenated_input_texts
        
        self.eval()
        show_progress_bar = True

        if convert_to_tensor:
            convert_to_numpy = False

        length_sorted_idx = np.argsort([-self._text_length(sen) for sen in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]
        all_embeddings = []

        if torch.cuda.device_count() <= 1:
            device = "cuda" if torch.cuda.is_available() else "cpu"
            self.to(device)
            for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
                sentences_batch = sentences_sorted[start_index:start_index + batch_size]
                embeddings = self._encode(sentences_batch, device=device, convert_to_numpy=convert_to_numpy)
                all_embeddings.append(embeddings)
        else:
            num_proc = torch.cuda.device_count()
            cuda_compatible_multiprocess = mp.get_context("spawn")
            with cuda_compatible_multiprocess.Pool(num_proc) as p:
                sentences_batches = [sentences_sorted[start_index:start_index + batch_size]
                                        for start_index in trange(0, len(sentences), batch_size)]
                for result in p.map(partial(self._encode,
                                            device=None,
                                            convert_to_numpy=convert_to_numpy,
                                            multiprocessing=True),
                                    sentences_batches):
                    all_embeddings.append(result)

        
        all_embeddings = torch.cat(all_embeddings, dim=0)
        all_embeddings = all_embeddings[np.argsort(length_sorted_idx)]
        all_embeddings = all_embeddings.to(torch.float32)
        if convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])
        return all_embeddings

    def save(self, output_path):
        self.model.save_pretrained(output_path)
        self.tokenizer.save_pretrained(output_path)

    def _encode(self, sentences_batch, device, convert_to_numpy, multiprocessing=False):

        if multiprocessing:
            rank = mp.current_process()._identity[0]
            if device is None and torch.cuda.is_available():
                device = f"cuda:{rank % torch.cuda.device_count()}"

        self.to(device)
        features = self.tokenize([self.prepare_for_tokenization(sentence) for sentence in sentences_batch])
        features = batch_to_device(features, device)

        with torch.no_grad():
            embeddings = self.forward(features)
            embeddings = embeddings.detach()
            embeddings = embeddings.cpu()

        return embeddings

    def _text_length(self, text: Union[List[int], List[List[int]]]):
        """
        Help function to get the length for the input text. Text can be either a string (which means a single text)
        a list of ints (which means a single tokenized text), or a tuple of list of ints
        (representing several text inputs to the model).
        """
        if isinstance(text, str) or (isinstance(text, list) and isinstance(text[0], int)) or len(text) == 0:     #Single text, list of ints, or empty
            return len(text)
        if isinstance(text, dict):                                                  #{key: value} case
            return len(next(iter(text.values())))
        elif not hasattr(text, '__len__'):                                          #Object has no len() method
            return 1
        else:
            return sum([len(t) for t in text])
        
    def resize_token_embeddings(
            self, new_num_tokens: Optional[int] = None, pad_to_multiple_of: Optional[int] = None
        ) -> nn.Embedding:
        return self.model.resize_token_embeddings(new_num_tokens=new_num_tokens, pad_to_multiple_of=pad_to_multiple_of)

==> llm2vec-0.1.2-py2.py3-none-any.whl: llm2vec/llm2vec.py <==
import logging
from functools import partial
import numpy as np
import torch
from torch import nn, Tensor, device
import torch.multiprocessing as mp
from typing import Dict, List, Union, Optional
from tqdm.autonotebook import trange
from transformers import AutoModel, AutoTokenizer, LlamaConfig, MistralConfig

logger = logging.getLogger(__name__)

def batch_to_device(batch, target_device: device):
    """
    send a pytorch batch to a device (CPU/GPU)
    """
    for key in batch:
        if isinstance(batch[key], Tensor):
            batch[key] = batch[key].to(target_device)
    return batch


class LLM2Vec(nn.Module):
    def __init__(
        self,
        model: AutoModel,
        tokenizer: AutoTokenizer,
        pooling_mode: str = "mean",
        max_length: int = 512,
        doc_max_length: int = 400,
        skip_instruction: bool = True
    ):
        super().__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.pooling_mode = pooling_mode
        self.skip_instruction = skip_instruction
        self.max_length = max_length
        self.doc_max_length = doc_max_length

    def prepare_for_tokenization(self, text):
        def _is_instruct(name):
            return ("chat" in name.lower()) or ("instruct" in name.lower()) or ("sharegpt" in name.lower())
        
        if _is_instruct(self.model.config._name_or_path):
            text = '[INST] ' + text.strip() + ' [/INST]'
        if (isinstance(self.model.config, LlamaConfig) or isinstance(self.model.config, MistralConfig)) and self.pooling_mode == "eos_token":
            text = text.strip() + ' </s>'
        return text

    def tokenize(self, texts):
        texts_2 = []
        original_texts = []
        for text in texts:
            t = text.split("!@#$%^&*()")
            texts_2.append(t[1])
            original_texts.append("".join(t))
        
        original = self.tokenizer(original_texts, return_tensors='pt', padding=True, truncation=True, max_length=self.max_length)
        embed_mask = None
        for t_i, t in enumerate(texts_2):
            ids = self.tokenizer([t], return_tensors='pt', padding=True, truncation=True, max_length=self.max_length, add_special_tokens=False)
            if embed_mask is None:
                e_m = torch.zeros_like(original["attention_mask"][t_i])
                if len(ids["input_ids"][0]) > 0:
                    e_m[-len(ids["input_ids"][0]):] = torch.ones(len(ids["input_ids"][0]))
                embed_mask = e_m.unsqueeze(0)
            else:
                e_m = torch.zeros_like(original["attention_mask"][t_i])
                if len(ids["input_ids"][0]) > 0:
                    e_m[-len(ids["input_ids"][0]):] = torch.ones(len(ids["input_ids"][0]))
                embed_mask = torch.cat((embed_mask, e_m.unsqueeze(0)), dim=0)

        original["embed_mask"] = embed_mask
        return original

    def _skip_instruction(self, sentence_feature):
        assert sentence_feature["attention_mask"].shape == sentence_feature["embed_mask"].shape
        sentence_feature["attention_mask"] = sentence_feature["embed_mask"]

    def forward(self, sentence_feature: Dict[str, Tensor]):
        embed_mask = None
        if "embed_mask" in sentence_feature:
            embed_mask = sentence_feature.pop("embed_mask")
        reps = self.model(**sentence_feature)
        sentence_feature["embed_mask"] = embed_mask
        
        return self.get_pooling(sentence_feature, reps.last_hidden_state)
    
    def get_pooling(self, features, last_hidden_states):  # All models padded from left
        assert self.tokenizer.padding_side == 'left', "Pooling modes are implemented for padding from left."
        if self.skip_instruction:
            self._skip_instruction(features)
        seq_lengths = features["attention_mask"].sum(dim=-1)
        if self.pooling_mode == "mean":
            return torch.stack([last_hidden_states[i, -length:, :].mean(dim=0) for i, length in enumerate(seq_lengths)], dim=0) 
        elif self.pooling_mode == "weighted_mean":
            bs, l, _ = last_hidden_states.shape
            complete_weights = torch.zeros(bs, l, device=last_hidden_states.device)
            for i, seq_l in enumerate(seq_lengths):
                if seq_l > 0:
                    complete_weights[i, -seq_l:] = torch.arange(seq_l) + 1
                    complete_weights[i] /= torch.clamp(complete_weights[i].sum(), min=1e-9)
            return torch.sum(last_hidden_states * complete_weights.unsqueeze(-1), dim=1)
        elif self.pooling_mode == "eos_token" or self.pooling_mode == "last_token":
            return last_hidden_states[:, -1]
        elif self.pooling_mode == "bos_token":
            return last_hidden_states[features["input_ids"]==self.tokenizer.bos_token_id]
        else:
            raise ValueError(f"{self.pooling_mode} is not implemented yet.")

    def _convert_to_str(self, instruction, text):   
        tokenized_q = self.tokenizer(text, return_tensors='pt', padding=True, truncation=True, max_length=self.max_length, add_special_tokens=False)
        tokenized_q_length = len(tokenized_q["input_ids"][0])

        while tokenized_q_length > self.doc_max_length:
            reduction_ratio = self.doc_max_length / tokenized_q_length
            reduced_length = int(len(text.split()) * reduction_ratio)
            text = " ".join(text.split()[:reduced_length])
            tokenized_q = self.tokenizer(text, return_tensors='pt', padding=True, truncation=True, max_length=self.max_length, add_special_tokens=False)
            tokenized_q_length = len(tokenized_q["input_ids"][0])
    
        return f"{instruction.strip()} !@#$%^&*(){text}"

    def encode(self, sentences: Union[str, List[str]],
               batch_size: int = 32,
               show_progress_bar: bool = True,
               convert_to_numpy: bool = False,
               convert_to_tensor: bool = False,
    ):
        if isinstance(sentences[0],str) and isinstance(sentences[-1],int):
            sentences = [sentences]
        # required for MEDI version of MTEB
        if isinstance(sentences[0],str):
            sentences = [[""] + [sentence] for sentence in sentences]
            
        concatenated_input_texts = []
        for sentence in sentences:
            assert isinstance(sentence[0], str)
            assert isinstance(sentence[1], str)
            concatenated_input_texts.append(self._convert_to_str(sentence[0], sentence[1]))
        sentences = concatenated_input_texts
        
        self.eval()
        show_progress_bar = True

        if convert_to_tensor:
            convert_to_numpy = False

        length_sorted_idx = np.argsort([-self._text_length(sen) for sen in sentences])
        sentences_sorted = [sentences[idx] for idx in length_sorted_idx]
        all_embeddings = []

        if torch.cuda.device_count() <= 1:
            device = "cuda" if torch.cuda.is_available() else "cpu"
            self.to(device)
            for start_index in trange(0, len(sentences), batch_size, desc="Batches", disable=not show_progress_bar):
                sentences_batch = sentences_sorted[start_index:start_index + batch_size]
                embeddings = self._encode(sentences_batch, device=device, convert_to_numpy=convert_to_numpy)
                all_embeddings.append(embeddings)
        else:
            num_proc = torch.cuda.device_count()
            cuda_compatible_multiprocess = mp.get_context("spawn")
            with cuda_compatible_multiprocess.Pool(num_proc) as p:
                sentences_batches = [sentences_sorted[start_index:start_index + batch_size]
                                        for start_index in trange(0, len(sentences), batch_size)]
                for result in p.map(partial(self._encode,
                                            device=None,
                                            convert_to_numpy=convert_to_numpy,
                                            multiprocessing=True),
                                    sentences_batches):
                    all_embeddings.append(result)

        
        all_embeddings = torch.cat(all_embeddings, dim=0)
        all_embeddings = all_embeddings[np.argsort(length_sorted_idx)]
        all_embeddings = all_embeddings.to(torch.float32)
        if convert_to_numpy:
            all_embeddings = np.asarray([emb.numpy() for emb in all_embeddings])
        return all_embeddings

    def save(self, output_path):
        self.model.save_pretrained(output_path)
        self.tokenizer.save_pretrained(output_path)

    def _encode(self, sentences_batch, device, convert_to_numpy, multiprocessing=False):

        if multiprocessing:
            rank = mp.current_process()._identity[0]
            if device is None and torch.cuda.is_available():
                device = f"cuda:{rank % torch.cuda.device_count()}"

        self.to(device)
        features = self.tokenize([self.prepare_for_tokenization(sentence) for sentence in sentences_batch])
        features = batch_to_device(features, device)

        with torch.no_grad():
            embeddings = self.forward(features)
            embeddings = embeddings.detach()
            embeddings = embeddings.cpu()

        return embeddings

    def _text_length(self, text: Union[List[int], List[List[int]]]):
        """
        Help function to get the length for the input text. Text can be either a string (which means a single text)
        a list of ints (which means a single tokenized text), or a tuple of list of ints
        (representing several text inputs to the model).
        """
        if isinstance(text, str) or (isinstance(text, list) and isinstance(text[0], int)) or len(text) == 0:     #Single text, list of ints, or empty
            return len(text)
        if isinstance(text, dict):                                                  #{key: value} case
            return len(next(iter(text.values())))
        elif not hasattr(text, '__len__'):                                          #Object has no len() method
            return 1
        else:
            return sum([len(t) for t in text])
        
    def resize_token_embeddings(
            self, new_num_tokens: Optional[int] = None, pad_to_multiple_of: Optional[int] = None
        ) -> nn.Embedding:
        return self.model.resize_token_embeddings(new_num_tokens=new_num_tokens, pad_to_multiple_of=pad_to_multiple_of)
